"""
Copyright 2010, 2011  Leif Theden


This file is part of lib2d.

lib2d is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lib2d is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lib2d.  If not, see <http://www.gnu.org/licenses/>.
"""


import sfx, gfx

# pygame related imports
import pygame
from pygame.transform import flip
from pygame.rect import Rect
from math import pi, ceil


pi2 = pi * 2

SOUND_PLAY  = 0
SOUND_HIT   = 1
SOUND_MISS  = 2
SOUND_BLOCK = 3


# there is limited tranformation support from pygame here.
# if this is greater than 0, then rotations will be
# limited to that number of angles in a 360 cicrle
simple_no = 128
simple_angle = 360.0 / simple_no


slow_pc = True


"""
TODO: make some sort of manager class for animations
      would make resource sharing less expensive
"""     

def padimage(image):    
    """
    Do a little processing of the input image to make it prettier.
    Pad the image with transparent pixels so the edges get antialised
    when rotated and scaled.
    """

    # replacement surface that is slightly bigger then the original
    new = Surface(image.get_rect().inflate(2, 2).size, pygame.SRCALPHA)

    # get the color on the border
    color = image.get_at((0,0))

    # set the color's alpha channel to be clear
    color[3] = 0

    # fill the replacement surface with a transparent background
    new.fill(color)

    new.blit(image, (1,1))

    return new


class AnimationFrame(object):
    """
    Animation Frames hold an image and a time value.  Add them to animations, please.

    The pygame rect object associated with the frame is only gaurentted to have
    correct size information.  frames should not be adjusted for position.  ever.
    """

    def __init__(self, image=None):
        if image == None:
            self.image = None
            self.rect = None
        else:
            self.image = image
            self.rect  = image.get_rect()

        # defaults
        self.ttl         = 1            # how many ms until flipped to next (-1 = never)
        self.axis        = (0,0)        # where images are aligned
        self.sound       = sfx.dummy    # sound played when frame is displayed

    def cleanup(self):
        self.image = None
        self.rect  = None
        self.sound = None

    # compute the rect's upper left corner when adjusted for the axis
    def adjust_position(self, position):
        new_pos = (position[0] - self.axis[0], position[1] - self.axis[1])
        return new_pos

    def load_image(self, filename):
        self.image, self.rect = gfx.load_image(filename)

    def load_sound(self, filename, sound_type=SOUND_PLAY):
        this_sound = sfx.load_sound(filename)
        self.sound = this_sound

    def __str__(self):
        return "AnimFrame \"%s\"" % id(self)

# TODO: make this more like a list object.  slices, etc.
class Animation(object):
    """
    Animation is a collection of frames with a few control variables.
    Animations can store different directions.

    directions:
        for the first type of rotation, sprites will be rotated in the center of each frame.

        for the second type of rotations (3d-like) you need different sets of animations.
            ideally, each direction would have the same number of frames, with same animation speed.
            each set of animation added will count as a seperate direction.
            1 animation  = no rotations
            2 animations = left and right (or up and down)
            4 animations = left, right, up, down
            6 animations = up, down, nw, ne, sw, se (for hex maps)
            and so on.

        ideally, you could have 4 directions for adventure games (even isometric)
        64 directions would be super sweet, but would require renderings of 3d models.
    """

    def __init__(self, name=None, dirs=1):
        self.name = name
        self.directions = dirs  # number of directions this animation supports (default is just "right")
        self.looping = 0        # if > 0, then loop x times.  -1 = forever
        self.loop_frame = 0     # if looping this is the start frame
        self.hold_play = False  # play and loop while button is held
        self.hold_loop = False  # loop while button is held, play at least once
        self.frames = ([],) * dirs

    def __len__(self):
        return self._len

    def cleanup(self):
        [ f.cleanup() for f in self.frames ]
        self.frames = None

    def get_frame(self, number, direction=0):
        """
        return the frame by number with the correct image for the direction
        """

        if direction < 0:
            direction = pi + (pi - abs(direction))

        d = int(ceil(direction / pi2 * (self.directions - 1)))

        return self.frames[d][number]

    def add_frame(self, frame, direction=0):
        """
        internally, frames is stored as a tuple to conserve memory.
        there is slight penalty when adding/removing frames since the
        internal "list" of frames must be converted to and from a tuple.

        if this is insane, and you have a reason to tell me why, let me know.

        direction is the direction this frames faces in radians.  if left blank, it is right.

        so, don't do anything tricky, like adding frames on the fly.  =)
        """

        frames = list(self.frames)
        tmp    = list(frames[direction])
        tmp.append(frame)
        frames[direction] = tuple(tmp)
        self.frames       = tuple(frames)
        self._len = len(self.frames[0])

    def __repr__(self):
        return "<Animation %s: \"%s\">" % (id(self), self.name)
