"""
Copyright 2010, 2011  Leif Theden


This file is part of lib2d.

lib2d is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lib2d is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lib2d.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
avatars are like sprites, but support multiple animations and advanced collision detection
an attached FSA will help control animations when controlled by player

====  N O T E S  ====

btw, i named these things avatars before that stupid movie.

avatars are not aware of the camera.  the basic avatar assumes the camera is "south" pointing "north".
this is the classical adventure style position for the camera.

i will possibly add camera support later for isometric games.

"""


from animation import Animation, AnimationFrame
from lib2d import DrawableObject, UpdateableObject, ObjectHandler
import gfx, sfx

import sys, weakref, pygame
from pygame.rect import Rect
from math import pi

DEBUG = False
#DEBUG = True

DRAW_COLLISION_BOXES = 0

def debug(text):
    if DEBUG: sys.stdout.write(text)


class PyGameSpriteMixin(object):
    """
    avatars can be compatable with pygame's dirtysprites

    the rect and image attributes are read only
    """

    @property
    def rect(self):
        return self.current_frame.rect

    @property
    def image(self):
        angle = self.parent.get_angle()
        if angle != self._old_angle:
            self.set_frame(self.current_frame_no, angle)

        try:
            return self.current_frame.image
        except AttributeError:
            print "no animation is playing!", self
            raise

    def get_visible(self):
        return self._is_visible

    def set_visible(self, value):
        self._is_visible = value

    # using properties, like a boss
    visible = property(get_visible, set_visible)


class Avatar(DrawableObject, UpdateableObject, PyGameSpriteMixin):
    """
    pygame compatable sprite that supports fsa, multiple animations and rotations

    an avatar must have a body associated with it
    an avatar must be connected to a renderer to be displayed

    update must be called occasionally for animations and rotations to work.
    see note in source code for details on the rotation system
    the fsa is a controller that manages animation changes and speed

    collision detection is not directly supported, although you maye make use of
    pygame's rect collision detection if you want to.  (you shouldnt want to)

    for a more fully-featured collision detection system, you can use a FighterAvatar
    which supports multiple collision boxes (hit boxes) for every frame.
    """

    def __init__(self, parent=None):
        self.default = None                 # what animation to play if not specified
        self.animations = {}                # the animations we can play
        self.loop_override       = False    # toggle force current animation loop
        self.loop_override_frame = None     # frame to loop if override applied
        self.callback = None                # this is called when animation is finished
        self.current_frame_no  = 0
        self.current_frame     = None
        self.current_animation = None
        self.previous_frame    = None
        self.callback = None
        self._is_paused = False
        self.looped = 0
        
        # timer initialization
        self.timer = 0

    @property
    def state(self):
        return (self.current_animation, self.current_frame_no)

    def update(self, time):
        """
        call this as often as possible with a time.  the units in the animation files
        must match the units provided here.  ie: milliseconds.

        bug: since timer is just normal python number, its possible to have a overflow.
        """

        if self._is_paused == True: return

        # this will happen if avatar is not given frames and such
        if self.current_frame == None: return

        self.timer += time
        while (self.timer >= self.current_frame.ttl) and self._is_paused == False:
            self.timer -= self.current_frame.ttl
            self.advance_frame()

    # sets the children of this to None, so that they can be garbage collected.
    # always call super last
    def cleanup(self):
        [ a.cleanup() for a in self.all_animations() ]

    def advance_frame(self):
        """
        advance the the frame in the current animation by one.

        animation will stop or loop here if needed.
        """

        if self._is_paused and self.current_frame.pause:
            return self._is_paused

        if self.current_animation == None:
            return

        self.current_frame_no += 1

        if self.current_frame_no >= len(self.current_animation):
            # the animation has reached the end
            if self.current_animation.looping == 0:
                return self.stop()

            # loop, but count the loops
            elif self.current_animation.looping > 0:
                if self.looped > self.current_animation.looping:
                    self.stop()
                else:
                    self.set_frame(self.current_animation.loop_frame)
                    self.looped += 1

            # loop forever, don't count the loops
            else:
                self.set_frame(self.current_animation.loop_frame)
        else:
            # just set the next frame
            self.set_frame(self.current_frame_no)

    def stop(self):
        """
        pauses the current animation and runs the callback if needed.
        """
        self.paused = True
        if self.callback != None:
            self.callback[0](*self.callback[1])

    @property
    def paused(self):
        return self._is_paused

    @paused.setter
    def paused(self, value):
        self._is_paused = value
        if value == False:
            self.previous_frame_no = self.current_frame_no

    def reset(self):
        """
        sets defaults of the avatar.
 
        the FSA attached may be locked sometimes during a reset.
        in that case, the play default will fail, since it won't play
        animations when locked.

        so, we need to unlock it, then relock the fsa if needed.
        """

        unlock = False

        try:
            if self.fsa.locked:
                unlock = True
        except:
            pass

        if self.default == None:
            self.current_animation = None
            self.current_frame_no = 0
            self.current_frame = None
            self.previous_frame = None
            self._is_paused = True
        else:
            if unlock:
                self.fsa.unlock()

            self.play(self.default)

            if unlock:
                self.fsa.lock()

    def set_frame(self, frame_no, angle=None):
        """
        set the frame of the animation

        """
        
        if angle == None:
            angle = self.parent.get_angle()
        self._old_angle = angle

        new_frame = self.current_animation.get_frame(frame_no, angle)

        self.previous_frame    = self.current_frame
        self.current_frame     = new_frame
        self.current_frame_no  = frame_no

        # ttl == -1: special way to stop an animation.
        if self.current_frame.ttl == -1:
            self.paused = True

    def play_norestart(self, *arg, **kwarg):
        """
        play an animation, but don't restart the current animation
        if the requested animation is the same
        """
        pass 
        

    def play(self, name=None, start_frame=0, callback=None, arg=[]):
        # reset some state stuff
        self.timer = 0
        self.looped = 0
        self.loop_override = False
        self.loop_override_frame = 0
        self.paused = False

        if callback != None:
            self.callback = (callback, arg)

        if name == None:
            name = self.default

        if isinstance(name, Animation):
            self.current_animation = name
        else:
            self.current_animation = self.get_animation(name)

        debug("playing %s: %s\n" % (self.current_animation, self))

        self.set_frame(start_frame)

    def force_loop(self, loop=None, frame_no=0):
        if loop != None: self.loop_override = loop
        self.loop_override_frame = frame_no
        return self.loop_override

    def add_animation(self, anim):
        # TODO: honor the animations.ini for the default animation

        if self.animations == {}:
            self.play(anim)

        self.animations[anim.name] = anim

    def get_animation(self, name):
        """
        return the animation for this name.
        """

        try:
            return self.animations[name]
        except KeyError:
            raise

    def set_default(self, name):
        """
        set the defualt animation to play if the avatar has nothing else to do
        """

        if isinstance(name, Animation):
            self.default = name
            return
        else:
            try:
                self.default = self.get_animation(name)
            except KeyError:
                return

    def __str__(self):
        return "<Avatar %s>" % id(self)


