"""
Copyright 2010, 2011  Leif Theden


This file is part of lib2d.

lib2d is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lib2d is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lib2d.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
load actor resources from sprite sheets

"""

from animation import Animation, AnimationFrame
from animation import SOUND_HIT, SOUND_BLOCK, SOUND_MISS, SOUND_PLAY
from avatar import Avatar
from rpg import Actor
import gfx

from configobj import ConfigObj
import os, os.path, sys, re, glob

from pygame import Rect
import pygame


DEBUG = False

# weeee, globals
g_resources   = "resources"
g_actors_root = g_resources + "/actors"
g_spritesheet = None
g_color_swap  = None
g_path = None

range_regex = re.compile(".*?(\d*)\s*-\s*(\d*)")


def debug(text):
    if DEBUG: sys.stdout.write(text)


class Spritesheet(object):
    """
    Simple class to manage spritesheets
    """

    def __init__(self, path, config):
        self.path   = path
        self.config = config
        self.image  = None
        self.rect   = None

    def load_image(self):
        try:
            filename = self.config['image']
        except KeyError:
            raise ValueError, "Cannot load image for spritesheet."

        # don't try to process the color key here.  when the spritesheet is cut up, we can handle it there.

        path = os.path.join(g_resources, "spritesheets", self.config['image'])
        self.image, self.rect = gfx.load_image(path)

class ActorLoader(object):
    """
    Load an actor.
    """

    def __init__(self, class_name=None, colorize=None):
        self.class_name = class_name
        self.colorize = colorize

def handle_bool(d, name, default=False):
    """
    Make parsing true/false a little easier
    """

    try:
        text = d[name].strip().lower()
    except KeyError:
        return default

    if text == "true":   return True
    if text == "yes":    return True
    if text == "no":     return False
    if text == "false":  return False

    raise ValueError


def load_actor(name, class_name=Actor, colorize=None):
    """
    load resources needed for an actor.

    return a actor.
    """

    global g_color_swap
    global g_path

    g_color_swap = colorize

    debug("loading actor...\n")

    actor = class_name()

    root            = os.path.join(g_actors_root, name)
    animations_path = os.path.join(root, "animations")

    g_path = root

    [ actor.avatar.add_animation(i) for i in load_animations(animations_path) ]

    #actor.avatar.reset()

    return actor

def load_animations(path):
    """
    load animations from a folder.

    each animation should be in one file.
    subdirectories will not be searched.
    """

    global g_spritesheet

    debug("loading animations:\n")

    a = []

    for name in glob.glob(path+'/*.ini'):
        debug("  \"%s\"\n" % name)
        a.append(read_animation_def(name))

    g_spritesheet = None

    return a

def read_general_options(d, animation, filename):
    """
    read the general section of the config.

    return a dict of options to add to each frame
    """

    global g_spritesheet

    general = {}

    try:
        d = d['general']
    except KeyError:
        return {}
    else:
        general.update(d)
        animation.hold_loop = handle_bool(d, 'hold loop')
        animation.hold_play = handle_bool(d, 'hold play')

        if handle_bool(d, "loop forever"):
            animation.looping = -1

        # if the animation is named, set it, otherwise use the filename
        try:
            animation.name = d['name']
        except KeyError:
            fn = os.path.split(filename)[-1]
            animation.name = fn[:fn.index(".")]
        
        # this animation uses a spritesheet, so
        # we are going to load the sheet now
        # so each frame doesnt load/unload the image
        try:
            filename = os.path.join(g_resources, "spritesheets", d['spritesheet']+".ini")
        except KeyError:
            pass
        else:
            config = ConfigObj(filename)
            g_spritesheet = Spritesheet(filename, config)
            g_spritesheet.load_image()

    return general

def load_grid_frame(animation, image, d):
    """
    a grid frame really is a set of images, one for each direction the
    animation supports.
    """

    w, h = read_point(d, 'dimension')
    ox, oy = read_point(d, 'offset')

    for y in range(int(d['directions'])):
        for x in range(int(d["frames"])):

            frame = AnimationFrame()
            frame.ttl   = float(d['ttl'])
            frame.hold  = handle_bool(d, 'hold', False)
            frame.pause = handle_bool(d, 'pause', True)

            # yes, i know pil can do conversions here.
            frame_image = pygame.surface.Surface((w, h))
            area = Rect((ox + (w * x), oy + (h * y)), (w, h))
            frame_image.blit(image, (0,0), area)

            try:
                if d['transparency'] == 'colorkey':
                    frame_image.set_colorkey(frame_image.get_at((0,0)))
                if d['transparency'] == 'pixel':
                    frame_image = frame_image.convert_alpha()
            except KeyError:
                pass

            if g_color_swap != None:
                new_image = frame_image.copy()
                pxarray = pygame.PixelArray(new_image)
                pxarray.replace(g_color_swap[0], g_color_swap[1])
                new_image.unlock()

                cached = gfx.get_image_from_cache(None, new_image)

                if cached == None:
                    gfx.add_image_to_cache(new_image)
                    frame_image = new_image
                else:
                    frame_image = cached

            frame.image = frame_image
            frame.rect = frame_image.get_rect()

            # add frame with direction information
            animation.add_frame(frame, y)

    return animation

def read_animation_grid(reader, filename):
    """
    animation grids are specially formatted spritesheets

    why?  i asked myself that too.  well, its just that most
    spritesheets you find online are formatted totally randomly.
    so, it would be nice to have one format for the loader,
    instead of trying to load every type.

    animation grids support multiple directions and are loaded
    *fairly* quickly.  nice, right?  they will also be nestable
    in the future, look out!

    these special spritesheets need to be created by the
    "cutup.py" script in the utilities folder.

    As opposed to individual images, sprites sheets have the
    advantages of being smaller and more easily compressed.
    """

    animation = Animation(dirs=int(reader["general"]["directions"]))
    general   = read_general_options(reader, animation, filename)

    fn = os.path.join(g_resources, "spritesheets", "grids", general["image"])
    image = pygame.image.load(fn)

    for section, d in reader.items():
        if section[:5].lower() == "frame":
            start = end = None
            d.update(general)

            try:
                start, end = range_regex.match(section).groups()
                start = int(start)
                end   = int(end)

            # AttributeError will be raised when it cannot parse a
            # start/end from the frame header.
            except AttributeError:
                # load a single frame
                d['name'] = section[6:]
                load_grid_frame(animation, image, d)

            else:
                # load a range of frames
                prefix, suffix = get_frame_image_filename_prefix_suffix(d)
                for x in range(start, end + 1):
                    # mangle the dict to reflect the right image file
                    d['file'] = prefix + str(x) + suffix
                    d['name'] = str(x)
                    load_grid_frame(animation, image, d)

    return animation


def read_animation_def(filename, frame_klass=AnimationFrame):
    """
    read an animation def file and return animation
    """

    reader      = ConfigObj(filename)
    path        = os.path.split(filename)[0]

    # animation grids require special handling
    if reader.has_key("general"):
        if reader["general"].has_key("layout"):
            if reader["general"]["layout"] == "animation grid":
                return read_animation_grid(reader, filename)

    animation = Animation()
    general   = read_general_options(reader, animation, filename)

    for section, d in reader.items():
        if section[:5].lower() == "frame":
            start = end = None
            d.update(general)

            try:
                start, end = range_regex.match(section).groups()
                start = int(start)
                end   = int(end)

            # AttributeError will be raised when it cannot parse a
            # start/end from the frame header.
            except AttributeError:
                # load a single frame
                d['name'] = section[6:]
                frame = load_frame(path, d, frame_klass)
                animation.add_frame(frame)

            else:
                # load a range of frames
                prefix, suffix = get_frame_image_filename_prefix_suffix(d)
                for x in range(start, end + 1):
                    # mangle the dict to reflect the right image file
                    d['file'] = prefix + str(x) + suffix
                    d['name'] = str(x)
                    frame = load_frame(path, d, frame_klass)
                    animation.add_frame(frame)    

    return animation


def read_point(d, name, default=(0,0)):
    try:
        return int(d[name][0]), int(d[name][1])
    except KeyError:
        return default
    except:
        raise ValueError, "Config Error: point %s improperly formatted" % name

def get_frame_image_filename_prefix_suffix(d):
    try:
        suffix = d['file suffix']
    except KeyError:
        suffix = ""

    try:
        prefix = d['file prefix']
    except KeyError:
        print "cannot find prefix for frame."
        print "you may ned to review the def.  may need to specify the file"
        raise

    return prefix, suffix

def get_frame_image_filename(d):
    try:
        image_filename = d['file']
    except KeyError:
        # DONT PANIC
        prefix, suffix = get_frame_image_filename_prefix_suffix(d)
        image_filename = prefix + d['name']

    return image_filename

def load_frame(path, d, klass=AnimationFrame):
    global g_color_swap

    sound_path = os.path.normpath(os.path.join(path, "../../sounds"))

    frame = klass()
    frame.ttl   = float(d['ttl'])
    frame.hold  = handle_bool(d, 'hold', False)
    frame.pause = handle_bool(d, 'pause', True)

    # frame has its own image
    if d.has_key('file'):
        image_filename = get_frame_image_filename(d)
        image_file = os.path.join(path, image_filename)
        try:
            if d['transparency'] == 'colorkey':
                frame.image, frame.rect = gfx.load_image_with_colorkey(image_file)
            if d['transparency'] == 'pixel':
                frame.image, frame.rect = gfx.load_image_with_alpha(image_file)
        except KeyError:
            frame.image, frame.rect = gfx.load_image_with_colorkey(image_file)

    # frame uses an image from a sprite sheet
    elif d.has_key('tile'):
        w, h = read_point(g_spritesheet.config, 'dimension')
        x, y = read_point(d, 'tile')
        image = pygame.surface.Surface((w, h))
        area = Rect((w * x, h * y), (w, h))
        image.blit(g_spritesheet.image, (0,0), area)

        try:
            if g_spritesheet.config['transparency'] == 'colorkey':
                image.set_colorkey(image.get_at((0,0)))
            if g_spritesheet.config['transparency'] == 'pixel':
                image = image.convert_alpha()
        except KeyError:
            pass

        frame.image = image
        frame.rect = image.get_rect()

    else:
        print "animation type not handled"
        raise Exception

    if g_color_swap != None:
        new_image = frame.image.copy()
        pxarray = pygame.PixelArray(new_image)
        pxarray.replace(g_color_swap[0], g_color_swap[1])
        new_image.unlock()

        cached = gfx.get_image_from_cache(None, new_image)

        if cached == None:
            gfx.add_image_to_cache(new_image)
            frame.image = new_image
        else:
            frame.image = cached

    try:
        if d['axis'] == 'center': frame.axis = frame.rect.center
    except KeyError:
        frame.axis = read_point(d, 'axis')

    if d.has_key('move'):
        frame.move_avatar = (int(d['move'][0]), int(d['move'][1]))

    # load collision boxes for the frame
    try:
        for x in range(1,6):
            section_name = "dmg box" + str(x)
            if d.has_key(section_name):
                hb = [int(x) for x in d[section_name]]
                box = CollisionBox(hb[:2], (hb[2]-hb[0]+1, hb[3]-hb[1]+1))
                frame.add_dmg_box(box)

        if d.has_key('block box'):
            hb = [int(x) for x in d['block box']]
            box = CollisionBox(hb[:2], (hb[2]-hb[0], hb[3]-hb[1]))
            frame.add_block_box(box)

        if d.has_key('hit box'):
            hb = [int(x) for x in d['hit box']]
            box = CollisionBox(hb[:2], (hb[2]-hb[0], hb[3]-hb[1]))
            frame.add_hit_box(box)

    except ValueError:
        print ""
        print "ERROR: improperly formatted hitbox"
        raise ValueError

    if d.has_key('block sound'):
        filename = os.path.join(sound_path, d['block sound'])
        frame.load_sound(filename, SOUND_BLOCK)

    if d.has_key('hit sound'):
        filename = os.path.join(sound_path, d['hit sound'])
        frame.load_sound(filename, SOUND_HIT)

    if d.has_key('miss sound'):
        filename = os.path.join(sound_path, d['miss sound'])
        frame.load_sound(filename, SOUND_MISS)

    if d.has_key('sound'):
        filename = os.path.join(sound_path, d['sound'])
        frame.load_sound(filename, SOUND_PLAY)

    for section_name in ("block spawn", "hit spawn"):
        if d.has_key(section_name):
            name       = d[section_name]
            extra_path = os.path.normpath(os.path.join(path, "../../extra"))
            anim_path  = os.path.join(extra_path, name, "anim.def")
            anim       = read_animation_def(anim_path, name)
            frame.__dict__[section_name.replace(" ","_")] = make_factory(anim)

    return frame

