"""
Copyright 2010, 2011  Leif Theden


This file is part of lib2d.

lib2d is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lib2d is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lib2d.  If not, see <http://www.gnu.org/licenses/>.
"""

import gfx
import pygame
from playerinput import KeyboardPlayerInput
from collections import deque
from pygame.locals import *


"""
pygame's font rendering is super slow, so rendering the fps
actually causes a measurable amount to slowdown when being
done each frame.

player's input doesn't get checked every loop.  it is checked
every 30ms and then handled.  this prevents the game logic
from dealing with input too often and slowing down the game.
"""


inputs = []
inputs.append(KeyboardPlayerInput())


def flush_cmds(cmds):
    pass


class StatePlaceholder(object):
    """
    holds a ref to a state

    when found in the queue, will be instanced
    """

    def __init__(self, klass):
        self.klass = klass

    def activate(self):
        pass

    def deactivate(self):
        pass

class StateDriver(object):
    """
    A state driver controls what is displayed and where input goes.

    A state is a logical way to break up "modes" of use for a game.
    For example, a title screen, options screen, normal play, pause,
    etc.
    """

    def __init__(self, parent=None):
        self.parent = parent
        self._states = deque()

        if parent != None:
            self.reload_screen()

    def get_size(self):
        return self.parent.get_screen().get_rect().size

    def get_screen(self):
        return self.parent.get_screen()

    def reload_screen(self):
        """ Called when the display changes mode somehow. """
        self._screen = self.parent.get_screen()

    def set_parent(self, parent):
        self.parent = parent
        self.reload_screen()

    def done(self):
        self.getCurrentState().deactivate()
        self._states.pop()
        state = self.getCurrentState()

        if isinstance(state, StatePlaceholder):
            self.replace(state.klass())
           
        elif state != None:
            if state.activated:
                state.reactivate()
            else:
                state.activate()

    def getCurrentState(self):
        try:
            return self._states[-1]
        except:
            return None

    def replace(self, state):
        self.getCurrentState().deactivate()
        self._states.pop()
        self.start(state)

    def start(self, state):
        self._states.append(state)
        self.getCurrentState().activate()

    def start_restart(self, state):
        """
        deserves an explanation

        start a new state, and like a normal start, when the
        new state finishes, then go back to the last state,
        BUT, the previous state will be feed first, then
        re-instanced (to cut down on memory usage)
        """

        prev = self.getCurrentState()
        prev.deactivate()
        self._states.pop()
        self._states.append(StatePlaceholder(prev.__class__))
        self.start(state)

    def push(self, state):
        self._states.appendleft(state)

    def run(self):
        currentState = self.getCurrentState()
        import gfx

        # deref for speed
        event_poll = pygame.event.poll
        event_pump = pygame.event.pump
        current_state = self.getCurrentState
        clock = pygame.time.Clock()

        #fps_font = pygame.font.Font("resources/fonts/northwoodhigh.ttf", 18)
        fps_font = pygame.font.Font(pygame.font.get_default_font(), 12)

        # streamline event processing by filtering out stuff we won't use
        pygame.event.set_allowed(None)
        pygame.event.set_allowed([QUIT, KEYDOWN, KEYUP, NOEVENT])

        # set an event to flush out commands
        event_flush = pygame.USEREVENT
        pygame.time.set_timer(pygame.USEREVENT, 30)

        # set an event to update the fps
        update_fps = pygame.USEREVENT + 1
        pygame.time.set_timer(pygame.USEREVENT + 1, 1000)
        fps = fps_font.render("fps: " + str(round(clock.get_fps(),0)), True, (200,200,0), (0,0,0)).convert()
        #fps = fps_font.render("fps: " + str(round(clock.get_fps(),0)), True, (200,200,0)).convert_alpha()

        # make sure our custom events will be triggered
        pygame.event.set_allowed([event_flush, update_fps])

        # some stuff to handle the command queue
        rawcmds = 0
        cmdlist = []
        checkedcmds = []
        
        while currentState:
            time = clock.tick(60)

            event = event_poll()
            while event.type != NOEVENT:

                # check each input for something interesting
                for cmd in [ c.getCommand(event) for c in inputs ]:
                    rawcmds += 1
                    if (cmd != None) and (cmd[:2] not in checkedcmds):
                        checkedcmds.append(cmd[:2])
                        cmdlist.append(cmd)

                # we should quit
                if event.type == QUIT:
                    currentState = None
                    break

                # do we update the fps?
                elif event.type == update_fps:
                    fps = fps_font.render("fps: " + str(round(clock.get_fps(),0)), True, (200,200,0), (0,0,0)).convert()

                # do we flush input now?
                elif event.type == pygame.USEREVENT:
                    currentState.handle_commandlist(cmdlist)
                    rawcmds = 0
                    checkedcmds = []
                    cmdlist = []

                # back out of this state, or send event to the state
                if event.type == KEYDOWN:
                    if event.key == K_ESCAPE:
                        self.done()
                        break
                    else:
                        currentState.handle_event(event)
                else:
                    currentState.handle_event(event)


                event = event_poll()

            if currentState:
                currentState.update(time)
                currentState.draw(self._screen)
                currentState = current_state()

                gfx.screen.blit(fps, (5,5))
                
                gfx.update_display()


# singleton type object
driver = StateDriver()
