"""
Copyright 2010, 2011  Leif Theden


This file is part of lib2d.

lib2d is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lib2d is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lib2d.  If not, see <http://www.gnu.org/licenses/>.
"""

from pygame.surface import Surface
from vec import Vec2d
from subpixelsurface import SubPixelSurface
from lib2d import UpdateableObject, DrawableObject
from statedriver import driver as sd
from utilities import base36encode as encode
import tiledtmxloader
import gfx

#from tiledtmxloader import tiledtmxloader

from math import ceil
from threading import Thread, Lock
from Queue import Queue as queue
from Queue import Empty
import sys, os.path

from pygame.rect import Rect
import pygame



# this image will be used when a tile cannot be loaded
def generate_default_image(size):
    i = Surface(size)
    i.fill((128, 128, 128))
    return i


class BigMapBlitThread(Thread):
    """
    Really, you only need one of these running.

    The map's buffer surface is locked automatically.
    ( really...only needed if you are using more than one
      thread, but nice to know it's locked anyway. )

    This thread will retreive tiles for BigMap objects
    and blit them to the parent as needed.

    This makes the impact of pygame's image reading/decoding
    less likely to affect the motion of the game.  nice.

    only tested on a 2010 macbook, os x lion.

    issues:
        seems like the constant creation and destruction of
        the drawing/blit thread causes a bit of jitter.
        could be better if thread is left alive for a while
        before being killed
    """

    def __init__(self, parent):
        Thread.__init__(self)
        self.parent = parent

    def run(self):
        # cache a couple objects to reduce attribute lookups
        toScreen  = self.parent.toScreen
        queue     = self.parent.queue
        blit      = self.parent.buf.blit
        get_tile  = self.parent.watch.get_tile
        buf_lock  = self.parent.buf_lock

        while 1:
            try:
                pos = queue.get(False)
            except Empty:
                break
            else:
                tile = get_tile(pos)
                pos = toScreen(pos)
                with buf_lock:
                    blit(tile, pos)
                queue.task_done()

class OneTileMap(object):
    """
    Class only accepts one image as a tile

    Used for simple parallax scrolling maps
    """

    def __init__(self, image):
        self.image = gfx.load(image)
        rect = self.image.get_rect()
        self.xtile, self.ytile = rect.size
        self.size = 1, 1
        self.layers = [None]

    def get_tile(self, p):
        return self.image

class TMXMap(object):
    """
    Map for Tiled TMX maps
    """

    def __init__(self, path):
        self.tmx = tiledtmxloader.TileMapParser().parse_decode(path)
        self.tmx.load(tiledtmxloader.ImageLoaderPygame())

        self.xtile = 16
        self.ytile = 16

        self.layers = self.tmx.layers
        self.size = (self.tmx.tilewidth, self.tmx.tileheight)
        self.default_image = generate_default_image((self.xtile, self.ytile))

    def get_tile(self, p):
        """
        loads tiles from images with encoded filenames.
        """

        if (p[0] < 1 ) or (p[1] < 1):
            return self.default_image

        try:
            gid = self.tmx.layers[p[2]].content2D[p[0]][p[1]]
            if gid == 0: return None
            tile, image = self.tmx.indexed_tiles[gid]
            return image
        except:
            return self.default_image

class BigMap(object):
    """
    This class loads a map as a collection of tiles.

    BigMap
        using a collection of images to use a map in-game.
        must be split using the "slice.py" utility.

    an idea for a proper "bigmap" file:

        * WHY???  so that tiles can be loaded without
                  loading into memory and decoding the entire image

        slice images
        each slice will be compressed (gzip?)
        record the current lenght of the file
        append each slice to a single file
        add the position to a list at the front of the file (toc)

        the header will have a name, author, source, tile size, map size
        the toc will be a list of byte locations where the tile starts
        the toc will be read when loaded, and stored as a dict
        to get a tile, lookup the byte location of the tile
        the data will be between that and the next tile

        read the data, unzip it, then read it as a pygame image.
        this will be a straight-forward way to store a map
        instead of using a ton of images, which is wasteful.

        with a little work, this could also be used to effecently
        render tileset type maps, such as one made with tiled

    """
    def __init__(self, location):
        self.location = location
        self.read_metadata()
        self.layers = [None]

        # placeholder image for missing tiles
        self.default_image = generate_default_image((self.xtile, self.ytile))

    def read_metadata(self):
        """
        read info about the map

        data file should be:
        prefix
        xsize ysize
        xtiles ytiles
        EOF

        prefix is the prefix of each file before the encoded coordinates
        x,ysize is the size of each tile in pixels
        x,ytiles is the number of tiles on each axis
        """

        try:
            with open(self.location + "map.txt") as fh:
                self.image_prefix = fh.readline().strip()
                self.xtile, self.ytile = tuple([int(i) for i in fh.readline().strip().split()])
                self.size = tuple([int(i) for i in fh.readline().strip().split()])
        except IOError:
            raise IOError, "cannot read metadata for map %s" % self.location

    def get_tile(self, p):
        """
        loads tiles from images with encoded filenames.
        """

        v = (p[0] << 8) | p[1]
        key = encode(v)
        try:
            filename = os.path.join(self.location, self.image_prefix) + key + ".png"
            return pygame.image.load(filename)
        except:
            return self.default_image

class TileMapRenderer(UpdateableObject, DrawableObject):
    """
    Render a TileMap object.  Render a map.  Yepp..

    If background update is set, then call update sometimes to render
    the layers in steps, rather than all at once.  yeee-ah.

    * subpixel rendering only works with OneTileMaps.
      it is too slow from use with multilayer and other maps.
    """

    def __init__(self, tilemap, size, subpixel=False):
        self.background_update = True
        self.subpixel = subpixel
        self.size = size

        # we keep it classy with colorkey blits for speed
        self.colorkey = (255,0,255)

        # hacks
        self.xtile = tilemap.xtile
        self.ytile = tilemap.ytile

        # more hacks
        tilemap.clear_image = Surface((self.xtile, self.ytile))
        tilemap.clear_image.fill(self.colorkey)
        #tilemap.clear_image.set_colorkey(self.colorkey)

        # start the camera in center
        self.xoffset = 0
        self.yoffset = 0

        # rect of the surface to manage
        self.rect = Rect(0,0, size[0]+self.xtile*2, size[1]+self.ytile*2)

        # rect of the tiles on the surface
        self.view = Rect(0,0, (size[0] / tilemap.xtile), \
                              (size[1] / tilemap.ytile))

        # self.layers is a list of surfaces
        self.layers = [ Surface(self.rect.size) for layer in tilemap.layers ]
        [ l.fill(self.colorkey) for l in self.layers[1:] ]
        [ l.set_colorkey(self.colorkey) for l in self.layers[1:] ]

        # hacks:
        self.get_tile = tilemap.get_tile
        self.queue = []
        self.tilemap = tilemap

    def adjust_view(self, v):
        """
        adjusts the view by re-tiling the background

        v is a tuple that expresses how the axis should be adjusted
        in tiles, not pixels.
        """

        # HACK:  subpixel rendering will only work with OneTileMaps
        if self.subpixel: return

        # make sure all of our old tiles are drawn before we continue
        self.flush_queue()

        # set our new view boundries
        self.view = self.view.move(v)

        # blitting the image over itself and scroling do the same thing
        # haven't benchmarked to see what is faster yet

        # copy the layer onto itself
        [ l.set_colorkey(None) for l in self.layers[1:] ]
        area = (self.toSurface(v), self.rect.size)
        [ l.blit(l, (0,0), area) for l in self.layers ]
        [ l.set_colorkey(self.colorkey) for l in self.layers[1:] ]

        # scroll the image
        #x, y = self.toSurface(v)    
        #[ layer.scroll(-x, -y) for layer in self.layers ]        

        # queue the missing tiles
        self.queue_edge_tiles(v)

        self.dirty = True

        if self.subpixel:
            self.sub = SubPixelSurface(self.buf, x_level=4)

    def queue_edge_tiles(self, v):
        """
        add the tiles on the edge that need to be redrawn
        """

        if v[0] > 0:
            # right
            self.queue.extend([
                (self.view.right + 1, y)
                for y in range(self.view.top, self.view.bottom + 2) ])
        elif v[0] < 0:
            # left
            self.queue.extend([
                (self.view.left, y)
                for y in range(self.view.top, self.view.bottom + 2) ])

        if v[1] > 0:
            # down
            self.queue.extend([
                (x, self.view.bottom + 1)
                for x in range(self.view.left, self.view.right + 2) ])
        elif v[1] < 0:
            # up
            self.queue.extend([
                (x, self.view.top)
                for x in range(self.view.left, self.view.right + 2) ])

    def update(self, time):
        """
        if background update is enabled, then you shuld call this often
        to make sure tiles along the edge are drawn.
        """

        if not self.background_update: return

        try:
            tile = self.queue.pop()
        except:
            return

        for i, layer in enumerate(self.layers):
            image = self.get_tile((tile[0], tile[1], i))
            layer.fill(self.colorkey, (self.toScreen(tile), (16,16)))
            if image != None: layer.blit(image, self.toScreen(tile))

        self.dirty = True

    def flush_queue(self):
        """
        draw any sprites that are sitting in the queue
        """
        
        for tile in self.queue:
            for i, layer in enumerate(self.layers):
                image = self.get_tile((tile[0], tile[1], i))
                layer.fill(self.colorkey, (self.toScreen(tile), (16,16)))
                if image != None: layer.blit(image, self.toScreen(tile))

        self.queue = []
        self.dirty = True

    def draw_layer(self, surface, layer):
        """
        layer should be a number
        """

        if self.subpixel:
            surface.blit(self.sub.at(-self.xoffset, -self.yoffset), (-self.xoffset,-self.yoffset))
        else:
            surface.blit(layer, (-self.xoffset,-self.yoffset))

    def draw_layer_no(self, surface, layer_no):
        surface.blit(self.layers[layer_no], (-self.xoffset,-self.yoffset)) 

    def draw(self, surface):
        """
        draw the map onto a surface.

        probably not a great idea to call directly, since you will not be able to
        layer objects properly.
        """

        [ surface.blit(layer, (-self.xoffset,-self.yoffset)) for layer in self.layers ]

    def redraw(self):
        """
        redraw the entire map (!!)

        will be slow, you've been warned
        """

        get_tile = self.get_tile
        ox, oy = self.view.topleft

        [ l.fill(self.colorkey) for l in self.layers ]

        for i, layer in enumerate(self.layers):
            for y in range(0, self.view.height + 2):
                for x in range(0, self.view.width + 2):
                    image = get_tile((x+ox, y+oy, i))
                    if image != None: layer.blit(image, self.toSurface((x,y)))

        if self.subpixel:
            self.sub = SubPixelSurface(self.surface, x_level=4)

    def scroll(self, v):
        """
        move the background in pixels
        """

        self.center((v[0] + self.xcenter, v[1] + self.ycenter))

    def center(self, pos):
        """
        center the map on a pixel
        """

        self.xcenter, self.ycenter = pos

        # calc the new postion in tiles and offset
        left, self.xoffset = divmod(pos[0] - self.size[0] / 2 , self.xtile)
        top, self.yoffset = divmod(pos[1] - self.size[1] / 2, self.ytile)

        # determine if tiles should be redrawn
        dx = left - self.view.left
        dy = top - self.view.top

        if (abs(dx) > 1) or (abs(dy) > 1):
            self.view.move_ip((dx, dy))
            self.redraw()

        elif (dx != 0) or (dy != 0):
            #self.view.move_ip((dx, dy))
            #self.redraw()
            self.adjust_view((dx, dy))

        else:
            #self.redraw()
            pass

        self.dirty = True

    def toScreen(self, p):
        """
        Adjusted for change in the view
        """

        return (p[0] * self.xtile - (self.view.left * self.xtile),
                p[1] * self.ytile - (self.view.top * self.ytile))

    def toSurface(self, p):
        """
        Convert map coordinates to surface/screen coords
        """

        return (p[0] * self.xtile,
                p[1] * self.ytile)

    def toMap(self, p):
        """
        Convert screen coordinate to map coords
        """

        return p[0] / self.xtile, p[1] / self.ytile

class ViewBase(UpdateableObject, DrawableObject):
    """
    Attach this class to a map to get a render of it.

    using this class is a bit different, you should do the following
    to render it:

    * call update on it if you want tiles to be rendered in background,
      rather than on demand.
    * call draw with a surface for a blit

    this order is pretty effiecent since you dont need to waste another
    buffer when setting the offsets.  also, internally, the buffer is
    never blitted until the offsets exceed the tile's size

    by default the view will render the map when it gets updated, rather
    than all at once.  as long as the map isn't scrolling very fast, then
    you won't notice it.  if the system can't keep up, you'll see black
    areas where the tiles were not loaded.

    """

    def __init__(self, watch, size, subpixel=False):
        self.subpixel = subpixel
        self.watch = watch
        self.background_update = True

        self.follow_obj = None

        # BUG: some bug prevents this from being larger than one
        edgeBuffer = 1

        # used to scroll the map
        self.xoffset = 0 
        self.yoffset = 0
        self.oldx = 0
        self.oldy = 0

        # the rect we occupy
        self.rect = Rect(0,0,size[0]+watch.xtile*(edgeBuffer*2), size[1]+watch.ytile*(edgeBuffer*2))

        self.bound = Rect(edgeBuffer*watch.xtile,
                          edgeBuffer*watch.ytile,
                          watch.size[0]+watch.xtile,
                          watch.size[1]+watch.ytile)

        # the tiles we are showing (also...a map)
        self.view = Rect((0,0), (size[0] / watch.xtile,
                                 size[1] / watch.ytile))

        # our surface to manage
        self.buf = Surface(self.rect.size).convert()

        # lazy
        self.xtile = watch.xtile
        self.ytile = watch.ytile
        self.size = self.watch.size

        self.xcenter = 0
        self.ycenter = 0
    
        self.dirty = False

    def adjust_view(self, v):
        """
        adjusts the view by re-tiling the background

        v is a tuple that expresses how the axis should be adjusted
        in tiles, not pixels.
        """

        new = self.view.move(v)
        # HACK:  subpixel rendering will only work with OneTileMaps
        if self.subpixel: return

        #if new.top < 0: return
        #if new.left < 0: return

        # set our new view boundries
        self.view = new

        # redraw any sprites that have not be drawn yet
        self.flush_queue()

        # blit the map onto itself, leaving the "new" portions open
        self.buf.blit(self.buf, (0,0), (self.toSurface(v), self.rect.size))

        # queue the missing tiles
        self.queue_edge_tiles(v)

        self.dirty = True

        if self.subpixel:
            self.sub = SubPixelSurface(self.buf, x_level=3)

    def flush_queue(self):
        """
        draw any sprites that are sitting in the queue
        """

        blit = self.buf.blit
        get  = self.watch.get_tile
        [ blit(get(i), self.toScreen(i)) for i in self.queue ]

    def queue_edge_tiles(self, v):
        """
        add the tiles on the edge that need to be redrawn
        """

        if v[0] > 0:
            # right
            self.queue.extend([
                (self.view.right + 1, y)
                for y in range(self.view.top, self.view.bottom + 2) ])
        elif v[0] < 0:
            # left
            self.queue.extend([
                (self.view.left, y)
                for y in range(self.view.top, self.view.bottom + 2) ])

        if v[1] > 0:
            # down
            self.queue.extend([
                (x, self.view.bottom + 1)
                for x in range(self.view.left, self.view.right + 2) ])
        elif v[1] < 0:
            # up
            self.queue.extend([
                (x, self.view.top)
                for x in range(self.view.left, self.view.right + 2) ])

    def redraw(self):
        """
        redraw the entire map

        will not blit tiles in the background, so use sparingly.
        """

        blit = self.buf.blit
        get_tile = self.watch.get_tile

        ox, oy = self.view.topleft

        for y in range(self.rect.top, self.rect.bottom / self.ytile):
            for x in range(self.rect.left, self.rect.right / self.xtile):
                blit(get_tile((x+ox,y+oy)), self.toSurface((x,y)))

        if self.subpixel:
            self.sub = SubPixelSurface(self.buf, x_level=3)

        self.dirty = True

    def follow(self, obj):
        """
        not tested
        """

        self.follow_obj = obj

    def scroll(self, v):
        """
        move the background in pixels
        """

        self.center((v[0] + self.xcenter, v[1] + self.ycenter))

    def center(self, pos):
        """
        center the map on a pixel
        """

        # BUG: if we don't redraw here, other layers get messed up
        if pos == (self.xcenter, self.ycenter):
            self.dirty = True
            return

        self.xcenter = pos[0]
        self.ycenter = pos[1]

        # calc the new postion in tiles and offset
        left, self.xoffset = divmod(pos[0], self.xtile)
        top, self.yoffset = divmod(pos[1], self.ytile)

        # determine if tiles should be redrawn
        dx = left - self.view.left
        dy = top - self.view.top

        if (dx != 0) or (dy != 0):
            self.adjust_view((dx, dy))
        else:
            self.dirty = True

    def toScreen(self, p):
        """
        Adjusted for change in the view
        """

        return (p[0] * self.xtile - (self.view.left * self.xtile),
                p[1] * self.ytile - (self.view.top * self.ytile))

    def toSurface(self, p):
        """
        Convert map coordinates to surface/screen coords
        """

        return (p[0] * self.xtile,
                p[1] * self.ytile)

    def toMap(self, p):
        """
        Convert screen coordinate to map coords
        """

        return p[0] / self.xtile, p[1] / self.ytile

    def update(self, time=None):
        #if self.follow_obj != None:
        #    self.center((self.follow_obj[0], self.follow_obj[1]))

        # non-threaded update of the buffer
        if self.background_update:
            try:
                i = self.queue.pop()
                self.buf.blit(self.watch.get_tile(i), self.toScreen(i))
                self.dirty = True
            except:
                pass

    def draw(self, surface):
        # BUG: when player hits the edge of the map, not setting the dirty
        #      flag here will cause the other layers to mess up

        if self.subpixel:
            surface.blit(self.sub.at(-self.xoffset, -self.yoffset), (-self.xoffset,-self.yoffset))
        else:
            surface.blit(self.buf, (-self.xoffset,-self.yoffset))

        #if self.dirty:
        #    surface.blit(self.buf, (-self.xoffset,-self.yoffset))
        #    self.dirty = False

class View(ViewBase):
    """
    View uses a normal list and doesn't do anything fancy
    """

    def __init__(self, watch, size, subpixel):
        ViewBase.__init__(self, watch, size, subpixel)
        self.queue = []
        self.redraw()

class ThreadedView(ViewBase):
    """
    This View uses a thread to blit onto the surface.  This is good
    for maps which will require a lot of disk access, but not a good
    idea for tile maps which are loaded into memory.

    the GIL in cpython makes thread use suitable to i/o intensive
    functions, but cpu locked functions will cause the scheduler
    to behave unpredictably.

    I/O is handled by a thread that loads the images and blits them
    to the surface.  hopefully, they it have finished loading/blitting
    by the time the camera moves to the edge.  if not, then you will
    have visable black areas on the edge of the map.
    They will go away soon.
    
    besides threading, it is will be possible to enlarge the buffer
    on the edge of the map's surface so there is extra time to
    render the map.

    BUG: sometimes, the wrong tiles are biltted to the map.
         not for certain why, but could be caused by the thread
         not being killed properly and leftover items in the queue.

    """
       
    def __init__(self, *args, **kwargs):
        ViewBase.__init__(self, *args, **kwargs)

        self.background_update = False

        self.buf_lock = Lock()
        self.queue = queue((self.rect.width / self.xtile) * (self.rect.height / self.ytile))

        self.redraw()
        self.thread = BigMapBlitThread(self)
        self.thread.start()

    def update(self, time):
        """
        since a thread handles all the drawing, we can simply pass here
        """

        pass

    def redraw(self):
        """
        "redraw" the map

        the map's queue is just populated with what needs to be drawn.
        Thread will draw that map.
        """

        blit = self.buf.blit
        for y in range(self.rect.top, self.rect.bottom / self.ytile):
            for x in range(self.rect.left, self.rect.right / self.xtile):
                self.queue.put((x,y))
        self.dirty = True

        if self.subpixel:
            self.sub = SubPixelSurface(self.buf, x_level=4)

    def start_thread(self):
        """
        starts a thread to drawn the map

        if a thread is already running, wont change anything
        """

        if self.thread.is_alive() == False:
            self.thread = BigMapBlitThread(self)
            self.thread.start()

    def flush_queue(self):
        """
        force the helper thread to blit the tiles in the queue
        """
        
        self.queue.join()
        self.dirty = True

    def queue_edge_tiles(self, v):
        """
        add the edge tiles to the queue for the thread to handle
        """

        if v[0] > 0:
            [ self.queue.put(pos) for pos in [
                (self.view.right + 1, y)
                for y in range(self.view.top, self.view.bottom + 2) ] ]

        elif v[0] < 0:
            [ self.queue.put(pos) for pos in [
                (self.view.left, y)
                for y in range(self.view.top, self.view.bottom + 2) ] ]

        if v[1] > 0:
            [ self.queue.put(pos) for pos in [
                (x, self.view.bottom + 1)
                for x in range(self.view.left, self.view.right + 2) ] ]

        elif v[1] < 0:
            [ self.queue.put(pos) for pos in [
                (x, self.view.top)
                for x in range(self.view.left, self.view.right + 2) ] ]

        self.start_thread()
        self.dirty = True
